"""Base classes."""


class RepositoryFacade:
    """Abstract base class for a Repository Facade.

    From the Wikipedia entry on "Facade pattern":
      'A facade is an object that provides a simplified interface to a
      larger body of code...'
      https://en.wikipedia.org/wiki/Facade_pattern
    In this case, we provide a single point of access for all Repository
    classes grouped in a conceptual unit, encapsulate the db connection,
    provide a commit() function for saving changes, and implement the magic
    methods __exit__ and __enter__ so this class is valid for use in a "with"
    statement.

    Collecting multiple repositories together might be viewed as an
    inefficiency: it is simple enough to initialize one Repository class as and
    when it is needed. Indeed, this is how I use repositories.

    The Facade comes in handy where we want to share database context between
    Repository classes.

    Implementation details for Repository Facade classes will differ with the
    database used. The intention is for a subclass to be defined for each
    such case. See MySQLRepositoryFacade and MongoRepositoryFacade, for example.
    """

    def __init__(self):
        """Create a new RepositoryFacade."""

    def __enter__(self):
        self.__init__()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Implementation note - see:
        https://stackoverflow.com/questions/22417323/
        how-do-enter-and-exit-work-in-python-decorator-classes
        """
        self.dispose()

    def dispose(self):
        """Close the connection for disposal of the RepositoryFacade."""
        raise NotImplementedError()


class Repository:
    """Abstract Repository class

    The docstrings here will indicate the intention of the functions and their
    arguments. Child classes can extend these in part, requiring new arguments.

    The base class remains agnostic as to return types as this will vary across
    databases.

    The exception is the exists() function, which is implemented here.
    """

    def __init__(self, **kwargs):
        self._kwargs = kwargs

    def __enter__(self):
        self.__init__(**self._kwargs)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Implementation note - see:
        https://stackoverflow.com/questions/22417323/
        how-do-enter-and-exit-work-in-python-decorator-classes
        """
        self.dispose()

    def add(self, item):
        """Add one or more items to the database.

        Args:
          items: an object or list of objects to add.
        """
        raise NotImplementedError()

    def all(self, projection=None):
        """Get all records in the database.

        Args:
          projection: List, optional, of attributes to project.
        """
        raise NotImplementedError()

    def commit(self):
        """Save changes to the database."""
        raise NotImplementedError()

    def count(self):
        """Get a count of how many records are in this table/collection."""
        raise NotImplementedError()

    def delete(self, items=None, **kwargs):
        """Delete item(s) from the database.

        May either specify an object or list of objects to delete, or
        keyword arguments for primary key identification.

        Args:
          items: an object or list of objects to delete.
          kwargs: can specify primary key attribute name(s) and value(s).
        """
        raise NotImplementedError()

    def dispose(self):
        """Dispose of this Repository."""
        raise NotImplementedError()

    def exists(self, **kwargs):
        """Check if a record exists.

        Pass the primary key values in as keyword arguments.

        Returns:
          Boolean indicating if the record exists.
        """
        # NOTE: to project here or not?
        return len(list(self.search(**kwargs))) > 0

    def get(self, expect=True, projection=None, **kwargs):
        """Get an item from the database.

        Primary key values need to be passed as keyword arguments.

        Arguments:
          expect: Bool, whether to throw an error if not found.
          projection: List, optional, of attributes to project.
          kwargs: can specify the primary key attribute name(s) and value(s).
        """
        raise NotImplementedError()

    def search(self, projection=None, **kwargs):
        """Search for records in the database.

        Specify the search attributes and values as keyword arguments.

        NOTE: This must return a generator or list of records - this is
        necessary for the default implementation of the exists() method. Any
        change to this means you should re-implement exists().

        Args:
          projection: List, optional, of attributes to project.
          kwargs: attribute name(s) and value(s) to search on.

        Returns:
          Generator or list of records.
        """
        raise NotImplementedError()



"""Objects for interfacing with SQL databases."""
from hsdbi import base
from hsdbi import errors
import importlib
import sqlalchemy as sa
from sqlalchemy import orm as saorm
from sqlalchemy import func
from sqlalchemy.dialects import mysql


# Static Functions


def create_sql_session(connection_string):
    """Create a SQLAlchemy session from a connection string.

    Args:
      connection_string: String.

    Returns:
      sqlalchemy.orm.session.Session object.
    """
    engine = sa.create_engine(connection_string)
    session = saorm.sessionmaker(bind=engine)()
    return session


def print_sql(query):
    print(query.statement.compile(dialect=mysql.dialect(),
                                  compile_kwargs={'literal_binds': True}))


# SQL Implementations


class SQLFacade(base.RepositoryFacade):
    """Facade for SQL repositories.

    Attributes:
      session: the sqlalchemy session wrapped by this facade. There is a
        question of whether to expose this. It was decided to expose it because
        it will enable flexibility since consumers can directly use it if
        convenient, providing better extensibility.
    """

    def __init__(self, connection_string):
        """Create a new RepositoryFacade.

        This will create the self._engine and self.session variables from
        the passed connection string. It also saves self._connection_string
        for reference.

        Args:
          connection_string: String, the connection string to the database.
        """
        super(SQLFacade, self).__init__()
        self._connection_string = connection_string
        self.session = create_sql_session(connection_string)

    def __enter__(self):
        self.__init__(self._connection_string)
        return self

    def commit(self):
        """Save changes to the database."""
        self.session.commit()

    def dispose(self):
        """Dispose of this class - close the database connection."""
        self.session.close()


class SQLRepository(base.Repository):
    """Generic wrapper for db access methods for a table."""

    def __init__(self, primary_keys, class_type, orm_module,
                 connection_string=None, session=None, **kwargs):
        """Create a new Repository.

        Args:
          class_type: Type, the type of the object this repository will handle.
            It should be one of the orm classes.
          orm_module: String, the name of the module containing orm classes
            that this Repository will work on. E.g. 'db.orm'.
          primary_keys: List of strings, identify the attribute names that
            represent the primary keys for this class.
          connection_string: String, optional, but must pass one of either
            connection_string or session.
          session: SQLAlchemy session object, optional, but must pass one of
            either connection_string or session.
        """
        super(SQLRepository, self).__init__(**kwargs)
        self._class_type = class_type
        self._orm_module = orm_module
        self._orm_module_key = orm_module.replace('.', '_')
        globals()[self._orm_module_key] = importlib.import_module(orm_module)
        self._primary_keys = primary_keys
        self._connection_string = connection_string
        self._session = session
        if connection_string:
            self._connection_string = connection_string
            self._session = create_sql_session(self._connection_string)
        elif session:
            self._session = session
        else:
            raise ValueError('You must pass either a session or '
                             'connection string.')

    def __enter__(self):
        self.__init__(self._primary_keys, self._class_type, self._orm_module,
                      self._connection_string, self._session, **self._kwargs)
        return self

    def add(self, items):
        """Add one or more items to the database.

        Args:
          items: one or more objects of the intended type; can be a list or
            a single object.
        """
        if isinstance(items, list):
            for item in items:
                self._session.add(item)
        else:
            self._session.add(items)

    def all(self, projection=None):
        """Retrieve all items of this kind from the database.

        Args:
          projection: List, optional, of attributes to project.

        Returns:
          List of items of the relevant type; or Tuple of values if projected.
        """
        query = self._session.query(self._class_type)
        if projection:
            query = self.project(query, projection)
        return query.all()

    def commit(self):
        """Commit changes to the database."""
        self._session.commit()

    def count(self):
        """Count the number of records in the table.

        Returns:
          Integer, the number of records in the table.
        """
        return self._session\
            .query(func.count(eval('%s.%s.%s' % (self._orm_module_key,
                                                 self._class_type.__name__,
                                                 self._primary_keys[0]))))\
            .scalar()

    def delete(self, items=None, **kwargs):
        """Delete item(s) from the database.

        Can either specify items directly as a single object of the expected
        type, or as a list of such objects; or specify a record to delete
        by primary key via keyword arguments.

        Args:
          items: one or more objects of the intended type; can be a list or
            a single object.
          kwargs: can specify the primary key name(s) and value(s).

        Raises:
          ValueError: if neither items nor keyword arguments are specified.
          ValueError: if items are not specified and any primary key is
            missing from the keyword arguments.
        """
        if not items and len(kwargs) == 0:
            raise ValueError('You must specify either items or kwargs.')
        if items:
            if isinstance(items, list):
                for item in items:
                    self._session.delete(item)
            else:
                self._session.delete(items)
        else:
            for pk in self._primary_keys:
                if pk not in kwargs.keys():
                    raise TypeError('Missing keyword argument: %s' % pk)
            self._session.delete(self.get(**kwargs))

    def delete_all_records(self):
        """Delete all records from this table.

        Returns:
          Integer, the number of records deleted.
        """
        return self._session.query(self._class_type).delete()

    def dispose(self):
        """Dispose of the database connection."""
        self._session.close()

    def get(self, expect=True, projection=None, debug=False, **kwargs):
        """Get an item from the database.

        Pass the primary key values in as keyword arguments.

        Args:
          expect: whether or not to expect the result. Will raise an exception
            if not found if True.
          projection: List of String attribute names to project, optional.
          kwargs: can specify the primary key attribute name(s) and value(s).
          debug: Boolean, if true will print the SQL of the query for
            debugging purposes.

        Raises:
          TypeError: if any primary key values are missing from the keyword
            arguments received.
          NotFoundError: if the item is not found in the database and the expect
            flag is set to True.
        """
        for pk in self._primary_keys:
            if pk not in kwargs.keys():
                raise TypeError('Missing keyword argument: %s' % pk)
        query = self._session.query(self._class_type)
        for attr, value in kwargs.items():
            query = query.filter(
                eval('%s.%s.%s == "%s"'
                     % (self._orm_module_key,
                        self._class_type.__name__, attr, value)))
        if projection:
            query = self.project(query, projection)
        if debug:
            print_sql(query)
        result = query.one_or_none()
        if not result and expect:
            raise errors.NotFoundError(pk=kwargs, table=self._class_type)
        if projection:
            return result[0]
        else:
            return result

    def project(self, query, projection, debug=False):
        """Perfoms a projection on the given query.

        Args:
          query: the query object to project.
          projection: List of String attributes to project.
          debug: Boolean, if true will print the SQL of the query for
            debugging purposes.

        Returns:
          Query object.

        Raises:
          ValueError if projection is not a list. It is easy to pass a string
            here, so we will catch this case for quick debugging.
        """
        if not isinstance(projection, list):
            raise ValueError('projection must be a list.')
        projection = ['%s.%s.%s' % (self._orm_module_key,
                                    self._class_type.__name__, attr)
                      for attr in projection]
        query = query.with_entities(eval(', '.join(projection)))
        if debug:
            print_sql(query)
        return query

    def search(self, projection=None, debug=False, **kwargs):
        """Attempt to get item(s) from the database.

        Pass whatever attributes you want as keyword arguments.

        Args:
          projection: List of String attribute names to project, optional.
          debug: Boolean, if true will print the SQL of the query for
            debugging purposes.

        Returns:
          List of matching results; or Tuple if projected.
        """
        query = self._session.query(self._class_type)
        for attr, value in kwargs.items():
            query = query.filter(
                eval('%s.%s.%s == "%s"'
                     % (self._orm_module_key,
                        self._class_type.__name__,
                        attr,
                        value)))
        if projection:
            query = self.project(query, projection)
        if debug:
            print_sql(query)
        return query.all()



from setuptools import setup, find_packages


# python setup.py sdist upload -r pypi


setup(
    name='hsdbi',
    packages=find_packages(exclude=['testing']),
    version='0.1a16',
    description='A simple interface for accessing databases.',
    author='Tim Niven',
    author_email='tim.niven.public@gmail.com',
    url='https://github.com/timniven/hsdbi',
    download_url='https://github.com/timniven/hsdbi/archive/0.1a16.tar.gz',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
    ],
    keywords='database interface facade',
    install_requires=[
        'pymongo',
        'sqlalchemy'
    ]
)



"""Custom errors for this package."""


class Error(Exception):
    pass


class NotFoundError(Exception):
    """A record was not found in the database.

    Attributes:
      pk: Dictionary, attr-value pairs representing the primary keys.
      table: the table searched.
    """

    def __init__(self, pk, table):
        self.pk = pk
        self.table = table



"""MongoDB implementations."""
from hsdbi import base
import pymongo
from hsdbi import errors


# Static Functions


def get_connection(server='localhost', port=27017):
    """Get a connection to a mongo server.

    Args:
      server: String, the address of the server, e.g. 'localhost' (the default).
      port: Integer, the port number to connect to, e.g. 27017 (the default).

    Returns:
      pymongo.mongo_client.MongoClient, connection to the database server.
    """
    return pymongo.MongoClient(host=server, port=port)


def projection_dict(projection):
    """Get a projection dictionary from a list of attribute names to project.

    Args:
      projection: List of string names of attributes to project.

    Returns:
      Dictionary like {'attr1': 1, 'attr': 1, ... }.
    """
    if projection:
        return dict(zip(projection, [1] * len(projection)))
    else:
        return {}


def sort(query, sort_key, sort_order):
    """Applies sort to an existing query.

    Args:
      query: pymongo.cursor.Cursor.
      sort_key: String.
      sort_order: String in {asc, desc}.

    Returns:
      pymongo.cursor.Cursor.
    """
    return query.sort(sort_key, pymongo.ASCENDING if sort_order == 'asc'
                                else pymongo.DESCENDING)


# MongoDB Implementations


class MongoFacade(base.RepositoryFacade):
    """"Meta" Facade implementation for MongoDB.

    With an SQL Repository facade, we connect to a single database instance.
    With a Mongo Repository facade however, we connect to the database server.
    So we will not specify a database to connect to. We leave this to the
    DBFacade class. This takes the connection from the MongoFacade class,
    creates a connection to the db required, and then passes this to the
    Repository class for further sub-connection to a collection. This will
    avoid the need to make multiple connections to the database server, and
    still provide a nice clean interface.

    Attributes:
      connection: pymongo.mongo_client.MongoClient, connection to the database
        server.
    """

    def __init__(self, server, port):
        """Create a new MongoRepositoryFacade.

        Args:
          server: String, the address of the server. E.g. 'localhost'.
          port: Integer, the port number to connect to. E.g. 27017.
        """
        super(MongoFacade, self).__init__()
        self._server = server
        self._port = port
        self.connection = get_connection(self._server, self._port)

    def __enter__(self):
        self.__init__(self._server, self._port)
        return self

    def __delitem__(self, key):
        pass

    def __getitem__(self, key):
        return self.__getattribute__(key)

    def __setitem__(self, key, value):
        pass

    def dispose(self):
        self.connection.close()


class MongoDbFacade:
    """Facade for a database in a mongo server.

    Attributes:
      db: pymongo.database.Database, the connection to the database.
    """

    def __init__(self, connection, db_name, collections=None):
        """Create a new MongoDbFacade.

        Args:
          connection: pymongo.mongo_client.MongoClient, connection to the
            database server.
          db_name: String, the name of the db to connect to.
          collections: optional list of Strings, the names of the collections.
            If specified MongoRepository objects will be automatically added
            to this class on initialization using these collection names. If
            this behaviour is not desired, leave collections as None, which is
            the default.
        """
        self._connection = connection
        self._db_name = db_name
        self._collections = collections
        self.db = connection.get_database(db_name)
        if collections:
            for collection_name in collections:
                exec('self.%s = '
                     'MongoRepository(self.db, collection_name)'
                     % collection_name)

    def __delitem__(self, key):
        pass

    def __enter__(self):
        self.__init__(self._connection, self._db_name, self._collections)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Implementation note - see:
        https://stackoverflow.com/questions/22417323/
        how-do-enter-and-exit-work-in-python-decorator-classes
        """
        self._connection.close()

    def __getitem__(self, key):
        return self.__getattribute__(key)

    def __setitem__(self, key, value):
        pass


class MongoRepository(base.Repository):
    """Repository implementation for MongoDB.

    If you override this class and add anything to the constructor arguments,
    you will need to also override the __enter__ method.
    """

    def __init__(self, db, collection_name):
        """Create a new MongoRepository.

        Args:
          db: pymongo.database.Database, the connection to the database.
          collection_name: String, the name of the collection we intend to
            connect to.
        """
        super(MongoRepository, self).__init__()
        self._db = db
        self._collection_name = collection_name
        self._collection = self._db.get_collection(self._collection_name)

    def __enter__(self):
        self.__init__(self._db, self._collection_name)
        return self

    def add(self, items=None, **kwargs):
        """Add one or more items to the database.

        Args:
          items: one or more objects of the intended type; can be a list or
            a single object.
          kwargs: if items is not specified, the kwargs dictionary is used
            to create the record.

        Raises:
          ValueError: if neither items nor kwargs are specified.
        """
        if items:
            if isinstance(items, list):
                for item in items:
                    self._collection.insert_one(item)
            else:
                self._collection.insert_one(items)
        elif len(kwargs) > 0:
            self._collection.insert_one(kwargs)
        else:
            raise ValueError('You must specify either items or kwargs')

    def all(self, projection=None, sort_key=None, sort_order='asc',
            batch_size=100):
        """Retrieve all items of this kind from the database.

        Args:
          projection: List, optional, of attributes to project.
          sort_key: String, optional.
          sort_order: String in {asc, desc}.
          batch_size: Integer, optional, how many records per get from the
            database server. Default is 100.

        Returns:
          pymongo.cursor.Cursor with results.
        """
        if projection:
            query = self._collection.find({}, projection_dict(projection))\
                                    .batch_size(batch_size)

        else:
            query = self._collection.find().batch_size(batch_size)
        if sort_key:
            query = sort(query, sort_key, sort_order)
        return query

    def commit(self):
        """Does nothing for a MongoRepository."""
        pass

    def count(self):
        """Count the number of records in the collection.

        Returns:
          Integer, the number of records in the collection.
        """
        return self._collection.count()

    def delete(self, items=None, **kwargs):
        """Delete item(s) from the database.

        Can either specify items directly as a single object of the expected
        type, or as a list of such objects; or specify a record to delete
        by keyword arguments.

        Args:
          items: one or more objects of the intended type; can be a list or
            a single object.
          kwargs: specify the attribute name(s) and value(s) of a document to
            be deleted.

        Raises:
          ValueError: if neither items nor keyword arguments are specified.
          ValueError: if items are not specified and any primary key is
            missing from the keyword arguments.
        """
        if not items and len(kwargs) == 0:
            raise ValueError('You must specify either items or kwargs.')
        if items:
            if isinstance(items, list):
                for item in items:
                    self._collection.delete_one(item)
            else:
                self._collection.delete_one(items)
        else:
            self._collection.delete_one(kwargs)

    def delete_all_records(self):
        """Delete all records from this collection."""
        self._collection.drop()

    def dispose(self):
        """Dispose of the database server connection."""
        self._db.client.close()

    def get(self, expect=True, projection=None, **kwargs):
        """Get an item from the database.

        Pass the primary key values in as keyword arguments.

        Args:
          expect: whether or not to expect the result. Will raise an exception
            if not found if True.
          projection: List of String attribute names to project, optional.
          kwargs: dictionary of search values.

        Returns:
          Dictionary: representing the retrieved document, if found.

        Raises:
          NotFoundError: if the item is not found in the database, but it is
            expected.
        """
        if projection:
            item = next(self._collection.find(kwargs,
                                              projection_dict(projection)),
                        None)
        else:
            item = next(self._collection.find(kwargs), None)
        if not item and expect:
            raise errors.NotFoundError(pk=kwargs, table=self._collection_name)
        return item

    def search(self, projection=None, sort_key=None, sort_order='asc',
               batch_size=100, **kwargs):
        """Attempt to get item(s) from the database.

        Pass whatever attributes you want as keyword arguments.

        Args:
          projection: List of String attribute names to project, optional.
          kwargs: the attribute name(s) and value(s) to be used for search.
            If empty this function is equivalent to all().
          sort_key: String.
          sort_order: String in {asc, desc}.
          batch_size: Integer, optional, how many records per get from the
            database server. Default is 100.

        Returns:
          pymongo.cursor.Cursor with matching results (if any).
        """
        if projection:
            query = self._collection.find(kwargs, projection_dict(projection))\
                                   .batch_size(batch_size)
        else:
            query = self._collection.find(kwargs).batch_size(batch_size)
        if sort_key:
            query = sort(query, sort_key, sort_order)
        return query

    def update(self, doc):
        """Update the doc, saving attribute states into the db.

        Args:
          doc: the document to update.
        """
        _id = doc['_id']
        doc.pop('_id')
        self._collection.update_one(
            {'_id': _id}, {'$set': doc})
        doc['_id'] = _id
